import pandas as pd
from datetime import datetime
from typing import Optional, List, Dict, Any, Union
from sqlalchemy.orm import Session
from logging import Logger
from requests import Response
from time import sleep
from math import nan
from stpstone._config.global_slots import YAML_MAIS_RETORNO_FUNDS
from stpstone.utils.cals.handling_dates import DatesBR
from stpstone.ingestion.abc.requests import ABCRequests
from stpstone.utils.webdriver_tools.playwright_wd import PlaywrightScraper
from stpstone.utils.parsers.dicts import HandlingDicts
from stpstone.utils.parsers.lists import ListHandler
from stpstone.utils.parsers.colors import ColorIdentifier
from stpstone.utils.parsers.folders import DirFilesManagement
from stpstone.utils.parsers.numbers import NumHandler


class MaisRetornoFunds(ABCRequests):

    def __init__(
        self,
        session: Optional[Session] = None,
        dt_ref: datetime = DatesBR().sub_working_days(DatesBR().curr_date, 1),
        cls_db: Optional[Session] = None,
        logger: Optional[Logger] = None,
        token: Optional[str] = None,
        list_slugs: Optional[List[str]] = None, 
        int_wait_load_seconds: int = 60, 
        int_delay_seconds: int = 30,
        bl_save_html: bool = False, 
        bl_headless: bool = False,
        bl_incognito: bool = False, 
        instruments_class: Optional[str] = None
    ) -> None:
        super().__init__(
            dict_metadata=YAML_MAIS_RETORNO_FUNDS,
            session=session,
            dt_ref=dt_ref,
            cls_db=cls_db,
            logger=logger,
            token=token,
            list_slugs=list_slugs,
            int_wait_load_seconds=int_wait_load_seconds, 
            int_delay_seconds=int_delay_seconds,
        )
        self.session = session
        self.dt_ref = dt_ref
        self.cls_db = cls_db
        self.logger = logger
        self.list_slugs = list_slugs
        self.int_wait_load_seconds = int_wait_load_seconds
        self.int_delay_seconds = int_delay_seconds
        self.bl_save_html = bl_save_html
        self.bl_headless = bl_headless
        self.bl_incognito = bl_incognito
        self.instruments_class = instruments_class
        self.list_months = [
            "JAN", "FEB", "MAR", "APR", "MAY", "JUN", "JUL", "AUG", "SEP", "OCT", "NOV", "DEC"
        ]

    def td_treatment_avl_funds(self, i: int, scraper: PlaywrightScraper) \
        -> Dict[str, Any]:
        p_cnpj = scraper.get_element(
            YAML_MAIS_RETORNO_FUNDS["avl_funds"]["xpaths"]["p_cnpj"].format(i),
            selector_type="xpath"
        )
        href_fund = scraper.get_element_attrb(
            YAML_MAIS_RETORNO_FUNDS["avl_funds"]["xpaths"]["href_fund"].format(i),
            str_attribute="href",
            selector_type="xpath"
        )
        a_fund_name = scraper.get_element(
            YAML_MAIS_RETORNO_FUNDS["avl_funds"]["xpaths"]["a_fund_name"].format(i),
            selector_type="xpath"
        )
        a_category = scraper.get_element(
            YAML_MAIS_RETORNO_FUNDS["avl_funds"]["xpaths"]["a_category"].format(i),
            selector_type="xpath"
        )
        status_fund = scraper.get_element(
            YAML_MAIS_RETORNO_FUNDS["avl_funds"]["xpaths"]["status_fund"].format(i),
            selector_type="xpath"
        )
        return {
            "CNPJ": p_cnpj.get("text", None),
            "URL_FUND": "https://maisretorno.com/" + href_fund if href_fund is not None \
                else None,
            "FUND_NAME": a_fund_name.get("text", None),
            "CATEGORY": a_category.get("text", None),
            "STATUS_FUND": status_fund.get("text", None),
            "PAGE_POSITION": i
        }
    
    def td_treatment_avl_instruments(self, i: int, scraper: PlaywrightScraper) \
        -> Dict[str, Any]:
        p_instrument_code = scraper.get_element(
            YAML_MAIS_RETORNO_FUNDS["avl_instruments"]["xpaths"]["p_instrument_code"].format(i),
            selector_type="xpath"
        )
        p_instrument_name = scraper.get_element(
            YAML_MAIS_RETORNO_FUNDS["avl_instruments"]["xpaths"]["p_instrument_name"].format(i),
            selector_type="xpath"
        )
        href_instrument = scraper.get_element_attrb(
            YAML_MAIS_RETORNO_FUNDS["avl_instruments"]["xpaths"]["href_instrument"].format(i),
            str_attribute="href",
            selector_type="xpath"
        )
        p_cnpj = scraper.get_element(
            YAML_MAIS_RETORNO_FUNDS["avl_instruments"]["xpaths"]["p_cnpj"].format(i),
            selector_type="xpath"
        )
        p_segment = scraper.get_element(
            YAML_MAIS_RETORNO_FUNDS["avl_instruments"]["xpaths"]["p_segment"].format(i),
            selector_type="xpath"
        )
        p_sector = scraper.get_element(
            YAML_MAIS_RETORNO_FUNDS["avl_instruments"]["xpaths"]["p_sector"].format(i),
            selector_type="xpath"
        )
        return {
            "CNPJ": p_cnpj.get("text", None),
            "INSTRUMENT_CODE": p_instrument_code.get("text", None),
            "INSTRUMENT_NAME": p_instrument_name.get("text", None),
            "URL_INSTRUMENT": "https://maisretorno.com/" + href_instrument if href_instrument is not None \
                else None,
            "SEGMENT": p_segment.get("text", None),
            "SECTOR": p_sector.get("text", None),
            "INSTRUMENTS_CLASS": self.instruments_class,
            "PAGE_POSITION": i
        }
    
    def td_treatment_avl_indexes(self, i: int, scraper: PlaywrightScraper) \
        -> Dict[str, Any]:
        p_index_name = scraper.get_element(
            YAML_MAIS_RETORNO_FUNDS["avl_indexes"]["xpaths"]["p_index_name"].format(i),
            selector_type="xpath"
        )
        p_index_code = scraper.get_element(
            YAML_MAIS_RETORNO_FUNDS["avl_indexes"]["xpaths"]["p_index_code"].format(i),
            selector_type="xpath"
        )
        href_index_url = scraper.get_element_attrb(
            YAML_MAIS_RETORNO_FUNDS["avl_indexes"]["xpaths"]["href_index_url"].format(i),
            str_attribute="href",
            selector_type="xpath"
        )
        p_status = scraper.get_element(
            YAML_MAIS_RETORNO_FUNDS["avl_indexes"]["xpaths"]["p_status"].format(i),
            selector_type="xpath"
        )
        return {
            "INDEX_NAME": p_index_name.get("text", None),
            "INDEX_CODE": p_index_code.get("text", None),
            "URL_INDEX": "https://maisretorno.com/" + href_index_url if href_index_url is not None \
                else None,
            "STATUS": p_status.get("text", None),
            "PAGE_POSITION": i
        }

    def td_treatment_fund_properties(self, scraper: PlaywrightScraper) -> List[Dict[str, Any]]:
        h1_fund_nickname = scraper.get_element(
            YAML_MAIS_RETORNO_FUNDS["fund_properties"]["xpaths"]["h1_fund_nickname"],
            selector_type="xpath"
        )
        p_fund_full_name = scraper.get_element(
            YAML_MAIS_RETORNO_FUNDS["fund_properties"]["xpaths"]["p_fund_full_name"],
            selector_type="xpath"
        )
        span_status = scraper.get_element(
            YAML_MAIS_RETORNO_FUNDS["fund_properties"]["xpaths"]["span_status"],
            selector_type="xpath"
        )
        hex_pool_open = scraper.get_element_attrb(
            YAML_MAIS_RETORNO_FUNDS["fund_properties"]["xpaths"]["svg_pool_open"],
            str_attribute="color",
            selector_type="xpath"
        )
        hex_qualified_investor = scraper.get_element_attrb(
            YAML_MAIS_RETORNO_FUNDS["fund_properties"]["xpaths"]["svg_qualified_investor"],
            str_attribute="color",
            selector_type="xpath"
        )
        hex_exclusive_fund = scraper.get_element_attrb(
            YAML_MAIS_RETORNO_FUNDS["fund_properties"]["xpaths"]["svg_exclusive_fund"],
            str_attribute="color",
            selector_type="xpath"
        )
        hex_long_term_taxation = scraper.get_element_attrb(
            YAML_MAIS_RETORNO_FUNDS["fund_properties"]["xpaths"]["svg_long_term_taxation"],
            str_attribute="color",
            selector_type="xpath"
        )
        hex_pension_fund = scraper.get_element_attrb(
            YAML_MAIS_RETORNO_FUNDS["fund_properties"]["xpaths"]["svg_pension_fund"],
            str_attribute="color",
            selector_type="xpath"
        )
        p_cnpj_fund = scraper.get_element(
            YAML_MAIS_RETORNO_FUNDS["fund_properties"]["xpaths"]["p_cnpj_fund"],
            selector_type="xpath"
        )
        p_benchmark = scraper.get_element(
            YAML_MAIS_RETORNO_FUNDS["fund_properties"]["xpaths"]["p_benchmark"],
            selector_type="xpath"
        )
        p_fund_initial_date = scraper.get_element(
            YAML_MAIS_RETORNO_FUNDS["fund_properties"]["xpaths"]["p_fund_initial_date"],
            selector_type="xpath"
        )
        p_fund_type = scraper.get_element(
            YAML_MAIS_RETORNO_FUNDS["fund_properties"]["xpaths"]["p_fund_type"],
            selector_type="xpath"
        )
        a_fund_administrator = scraper.get_element(
            YAML_MAIS_RETORNO_FUNDS["fund_properties"]["xpaths"]["a_fund_administrator"],
            selector_type="xpath"
        )
        a_fund_class = scraper.get_element(
            YAML_MAIS_RETORNO_FUNDS["fund_properties"]["xpaths"]["a_fund_class"],
            selector_type="xpath"
        )
        a_fund_subclass = scraper.get_element(
            YAML_MAIS_RETORNO_FUNDS["fund_properties"]["xpaths"]["a_fund_subclass"],
            selector_type="xpath"
        )
        a_fund_manager = scraper.get_element(
            YAML_MAIS_RETORNO_FUNDS["fund_properties"]["xpaths"]["a_fund_manager"],
            selector_type="xpath"
        )
        p_rentability_ltm = scraper.get_element(
            YAML_MAIS_RETORNO_FUNDS["fund_properties"]["xpaths"]["p_rentability_ltm"],
            selector_type="xpath"
        )
        p_aum = scraper.get_element(
            YAML_MAIS_RETORNO_FUNDS["fund_properties"]["xpaths"]["p_aum"],
            selector_type="xpath"
        )
        p_average_aum_ltm = scraper.get_element(
            YAML_MAIS_RETORNO_FUNDS["fund_properties"]["xpaths"]["p_average_aum_ltm"],
            selector_type="xpath"
        )
        p_volatility_ltm = scraper.get_element(
            YAML_MAIS_RETORNO_FUNDS["fund_properties"]["xpaths"]["p_volatility_ltm"],
            selector_type="xpath"
        )
        p_sharpe_ltm = scraper.get_element(
            YAML_MAIS_RETORNO_FUNDS["fund_properties"]["xpaths"]["p_sharpe_ltm"],
            selector_type="xpath"
        )
        p_qty_unitholders = scraper.get_element(
            YAML_MAIS_RETORNO_FUNDS["fund_properties"]["xpaths"]["p_qty_unitholders"],
            selector_type="xpath"
        )
        h3_manager_full_name = scraper.get_element(
            YAML_MAIS_RETORNO_FUNDS["fund_properties"]["xpaths"]["h3_manager_full_name"],
            selector_type="xpath"
        )
        p_fund_manager_director = scraper.get_element(
            YAML_MAIS_RETORNO_FUNDS["fund_properties"]["xpaths"]["p_fund_manager_director"],
            selector_type="xpath"
        )
        p_fund_manager_email = scraper.get_element(
            YAML_MAIS_RETORNO_FUNDS["fund_properties"]["xpaths"]["p_fund_manager_email"],
            selector_type="xpath"
        )
        p_fund_manager_site = scraper.get_element(
            YAML_MAIS_RETORNO_FUNDS["fund_properties"]["xpaths"]["p_fund_manager_site"],
            selector_type="xpath"
        )
        p_fund_manager_telephone = scraper.get_element(
            YAML_MAIS_RETORNO_FUNDS["fund_properties"]["xpaths"]["p_fund_manager_telephone"],
            selector_type="xpath"
        )
        p_minimum_investment_amount = scraper.get_element(
            YAML_MAIS_RETORNO_FUNDS["fund_properties"]["xpaths"]["p_minimum_investment_amount"],
            selector_type="xpath"
        )
        p_minimum_balance_required = scraper.get_element(
            YAML_MAIS_RETORNO_FUNDS["fund_properties"]["xpaths"]["p_minimum_balance_required"],
            selector_type="xpath"
        )
        p_administration_fee = scraper.get_element(
            YAML_MAIS_RETORNO_FUNDS["fund_properties"]["xpaths"]["p_administration_fee"],
            selector_type="xpath"
        )
        p_administration_fee_max = scraper.get_element(
            YAML_MAIS_RETORNO_FUNDS["fund_properties"]["xpaths"]["p_administration_fee_max"],
            selector_type="xpath"
        )
        p_performance_fee = scraper.get_element(
            YAML_MAIS_RETORNO_FUNDS["fund_properties"]["xpaths"]["p_performance_fee"],
            selector_type="xpath"
        )
        p_fund_quotation_period = scraper.get_element(
            YAML_MAIS_RETORNO_FUNDS["fund_properties"]["xpaths"]["p_fund_quotation_period"],
            selector_type="xpath"
        )
        p_fund_settlement_period = scraper.get_element(
            YAML_MAIS_RETORNO_FUNDS["fund_properties"]["xpaths"]["p_fund_settlement_period"],
            selector_type="xpath"
        )
        return [
            {
                "NICKNAME": h1_fund_nickname.get("text", None),
                "FUND_NAME": p_fund_full_name.get("text", None), 
                "STATUS": span_status.get("text", None),
                "BL_POOL_OPEN": ColorIdentifier(hex_pool_open).bl_green() \
                    if hex_pool_open is not None else False, 
                "BL_QUALIFIED_INVESTOR": ColorIdentifier(hex_qualified_investor).bl_green() \
                    if hex_qualified_investor is not None else False, 
                "BL_EXCLUSIVE_FUND": ColorIdentifier(hex_exclusive_fund).bl_green() \
                    if hex_exclusive_fund is not None else False, 
                "BL_LONG_TERM_TAXATION": ColorIdentifier(hex_long_term_taxation).bl_green() \
                    if hex_long_term_taxation is not None else False, 
                "BL_PENSION_FUND": ColorIdentifier(hex_pension_fund).bl_green() \
                    if hex_pension_fund is not None else False,
                "CNPJ": p_cnpj_fund.get("text", None), 
                "BENCHMARK": p_benchmark.get("text", None), 
                "FUND_INITIAL_DATE": DatesBR().str_date_to_datetime(
                    p_fund_initial_date.get("text", None), "DD/MM/YYYY"), 
                "FUND_TYPE": p_fund_type.get("text", None), 
                "ADMINISTRATOR": a_fund_administrator.get("text", None), 
                "CLASS": a_fund_class.get("text", None), 
                "SUBCLASS": a_fund_subclass.get("text", None), 
                "MANAGER": a_fund_manager.get("text", None) if a_fund_manager is not None \
                    else None, 
                "RENTABILITY_LTM": float(p_rentability_ltm.get("text", None).replace(
                    "%", "").replace(",", ".")) / 100.0 \
                    if p_rentability_ltm is not None and \
                        p_rentability_ltm.get("text", None) is not None and \
                        p_rentability_ltm.get("text", None) != "-" else None,
                "AUM": p_aum.get("text", None), 
                "AVERAGE_AUM_LTM": p_average_aum_ltm.get("text", None), 
                "VOLATILITY_LTM": float(p_volatility_ltm.get("text", None).replace(
                    "%", "").replace(",", ".")) / 100.0 \
                    if p_volatility_ltm.get("text", None) is not None and \
                        p_volatility_ltm.get("text", None) != "-" else None, 
                "SHARPE_LTM": float(p_sharpe_ltm.get("text", None).replace(",", ".")) \
                    if p_sharpe_ltm.get("text", None) is not None and \
                        p_sharpe_ltm.get("text", None) != "-" else None,
                "QTY_UNITHOLDERS": p_qty_unitholders.get("text", None), 
                "MANAGER_FULL_NAME": h3_manager_full_name.get("text", None), 
                "MANAGER_DIRECTOR": p_fund_manager_director.get("text", None), 
                "MANAGER_EMAIL": p_fund_manager_email.get("text", None), 
                "MANAGER_SITE": p_fund_manager_site.get("text", None), 
                "MANAGER_TELEPHONE": p_fund_manager_telephone.get("text", None),
                "MINIMUM_INVESTMENT_AMOUNT": float(p_minimum_investment_amount.get(
                    "text", None).replace(".", "").replace("R$ ", "")) \
                    if p_minimum_investment_amount.get("text", None) is not None \
                    and p_minimum_investment_amount.get("text", None) != "-" \
                    and len(p_minimum_investment_amount.get("text", None)) > 1 else None, 
                "MINIMUM_BALANCE_REQUIRED": float(p_minimum_balance_required.get(
                    "text", None).replace(".", "").replace("R$ ", "")) \
                    if p_minimum_balance_required.get("text", None) is not None \
                    and p_minimum_balance_required.get("text", None) != "-" \
                    and len(p_minimum_balance_required.get("text", None)) > 1 else None, 
                "ADMINISTRATION_FEE": float(p_administration_fee.get("text", None).replace(
                    "%", "").replace(",", ".")) / 100.0 \
                    if p_administration_fee.get("text", None) is not None \
                        and p_administration_fee.get("text", None) != "-" else None,
                "ADMINISTRATION_FEE_MAX": float(p_administration_fee_max.get("text", None).replace(
                    "%", "").replace(",", ".")) / 100.0 \
                    if p_administration_fee_max.get("text", None) is not None \
                        and p_administration_fee.get("text", None) != "-" else None,
                "PERFORMANCE_FEE": float(p_performance_fee.get("text", None).replace(
                    "%", "").replace(",", ".")) / 100.0 \
                    if p_performance_fee.get("text", None) is not None \
                        and p_performance_fee.get("text", None) != "-" else None,
                "FUND_QUOTATION_PERIOD": p_fund_quotation_period.get("text", None) \
                    if p_fund_quotation_period is not None \
                        and p_fund_quotation_period != "-" else None, 
                "FUND_SETTLEMENT_PERIOD": p_fund_settlement_period.get("text", None) \
                    if p_fund_settlement_period is not None \
                        and p_fund_settlement_period != "-" else None
            }
        ]
    
    def _convert_nums(self, list_: List[Union[str, float, int]], 
                      str_instrument: Optional[str] = None) -> List[Union[str, float, int]]:
        list_ = [NumHandler().transform_to_float(
                dict_, int_precision=6) for dict_ in list_]
        if str_instrument:
            list_ = [nan if x == "-" else str_instrument.upper() + " " + x if type(x) == str \
                    and "p.p." in x else x for x in list_]
        else:
            list_ = [nan if x == "-" else x for x in list_]
        return list_

    def td_instruments_historical_rentability(self, scraper: PlaywrightScraper) \
        -> List[Dict[str, Any]]:
        list_cols = ListHandler().extend_lists(["INSTRUMENT"], self.list_months, 
                                               ["YTD", "SINCE_INCEPTION"])
        str_instrument = DirFilesManagement().get_filename_parts_from_url(
            scraper.get_current_url())[0]
        list_years = scraper.get_elements(
            YAML_MAIS_RETORNO_FUNDS["instruments_historical_rentability"]["xpaths"]["list_years"],
            selector_type="xpath"
        )
        list_years = [int(dict_["text"]) for dict_ in list_years]
        list_years = list_years * 2
        list_td_rentabilities = scraper.get_elements(
            YAML_MAIS_RETORNO_FUNDS["instruments_historical_rentability"]["xpaths"][
                "list_td_rentabilities"],
            selector_type="xpath"
        )
        list_td_alpha = scraper.get_elements(
            YAML_MAIS_RETORNO_FUNDS["instruments_historical_rentability"]["xpaths"]["list_td_alpha"],
            selector_type="xpath"
        )
        list_td_rentabilities = self._convert_nums(list_td_rentabilities, str_instrument)
        list_td_alpha = self._convert_nums(list_td_alpha, str_instrument)
        list_ = ListHandler().extend_lists(list_td_rentabilities, list_td_alpha, 
                                           bl_remove_duplicates=False)
        list_ser = HandlingDicts().pair_headers_with_data(list_cols, list_)
        df_ = pd.DataFrame(list_ser)
        df_["YEAR"] = list_years
        return df_.to_dict(orient="records")
    
    def td_instruments_stats(self, scraper: PlaywrightScraper) -> List[Dict[str, Any]]:
        list_cols = ["YTD", "MTD", "LTM", "L_24M", "L_36M", "L_48M", "L_60M", "SINCE_INCEPTION"]
        list_stas = scraper.get_elements(
            YAML_MAIS_RETORNO_FUNDS["instruments_stats"]["xpaths"]["list_stas"],
            selector_type="xpath"
        )
        list_td = scraper.get_elements(
            YAML_MAIS_RETORNO_FUNDS["instruments_stats"]["xpaths"]["list_td"],
            selector_type="xpath"
        )
        list_td = self._convert_nums(list_td, "")
        list_ser = HandlingDicts().pair_headers_with_data(list_cols, list_td)
        df_ = pd.DataFrame(list_ser)
        df_["STATISTIC"] = [dict_["text"] for dict_ in list_stas]
        df_["INSTRUMENT"] = DirFilesManagement().get_filename_parts_from_url(
            scraper.get_current_url())[0].upper()
        list_ser = df_.to_dict(orient="records")
        list_cols = ["INSTRUMENT", "STATISTIC"] + list_cols
        df_ = df_[list_cols]
        return df_.to_dict(orient="records")
    
    def td_consistency(self, scraper: PlaywrightScraper) -> List[Dict[str, Any]]:
        span_positive_months = scraper.get_element(
            YAML_MAIS_RETORNO_FUNDS["instruments_consistency"]["xpaths"]["span_positive_months"],
            selector_type="xpath"
        )["text"]
        span_negative_months = scraper.get_element(
            YAML_MAIS_RETORNO_FUNDS["instruments_consistency"]["xpaths"]["span_negative_months"],
            selector_type="xpath"
        )["text"]
        span_greatest_return = scraper.get_element(
            YAML_MAIS_RETORNO_FUNDS["instruments_consistency"]["xpaths"]["span_greatest_return"],
            selector_type="xpath"
        )["text"]
        span_least_return = scraper.get_element(
            YAML_MAIS_RETORNO_FUNDS["instruments_consistency"]["xpaths"]["span_least_return"],
            selector_type="xpath"
        )["text"]
        return [{
            "INSTRUMENT": DirFilesManagement().get_filename_parts_from_url(
                scraper.get_current_url())[0].upper(),
            "POSITIVE_MONTHS": NumHandler().transform_to_float(span_positive_months, int_precision=6),
            "NEGATIVE_MONTHS": NumHandler().transform_to_float(span_negative_months, int_precision=6),
            "GREATEST_RETURN": NumHandler().transform_to_float(span_greatest_return, int_precision=6),
            "LEAST_RETURN": NumHandler().transform_to_float(span_least_return, int_precision=6)
        }]

    def req_trt_injection(self, resp_req: Response) -> Optional[pd.DataFrame]:
        i = 1
        list_ser = list()
        source = self.get_query_params(resp_req.url, "source")
        scraper = PlaywrightScraper(
            bl_headless=self.bl_headless,
            int_default_timeout=self.int_wait_load_seconds * 1_000, 
            bl_incognito=self.bl_incognito
        )
        with scraper.launch():
            if scraper.navigate(resp_req.url):
                if self.bl_save_html:
                    scraper.export_html(
                        scraper.page.content(), 
                        folder_path="data", 
                        filename="html-mais-retorno-avl-funds", 
                        bl_include_timestamp=True
                    )
                if source == "avl_funds":
                    while True:
                        if not scraper.selector_exists(
                            YAML_MAIS_RETORNO_FUNDS[source]["xpaths"]["p_cnpj"].format(i),
                            selector_type="xpath",
                            timeout=self.int_wait_load_seconds * 1_000
                        ): break
                        list_ser.append(self.td_treatment_avl_funds(i, scraper))
                        i += 1
                elif source == "avl_instruments":
                    while True:
                        if not scraper.selector_exists(
                            YAML_MAIS_RETORNO_FUNDS[source]["xpaths"]["p_instrument_code"].format(i),
                            selector_type="xpath",
                            timeout=self.int_wait_load_seconds * 1_000
                        ): break
                        list_ser.append(self.td_treatment_avl_instruments(i, scraper))
                        i += 1
                elif source == "avl_indexes":
                    while True:
                        if not scraper.selector_exists(
                            YAML_MAIS_RETORNO_FUNDS[source]["xpaths"]["p_index_code"].format(i),
                            selector_type="xpath",
                            timeout=self.int_wait_load_seconds * 1_000
                        ): break
                        list_ser.append(self.td_treatment_avl_indexes(i, scraper))
                        i += 1
                elif source == "fund_properties":
                    list_ser = self.td_treatment_fund_properties(scraper)
                elif source == "instruments_historical_rentability":
                    list_ser = self.td_instruments_historical_rentability(scraper)
                elif source == "instruments_stats":
                    list_ser = self.td_instruments_stats(scraper)
                elif source == "instruments_consistency":
                    list_ser = self.td_consistency(scraper)
                else:
                    raise Exception(f"Invalid source {source}. Please choose a valid one.")
        return pd.DataFrame(list_ser)


import pandas as pd
from datetime import datetime
from typing import Optional, List
from sqlalchemy.orm import Session
from logging import Logger
from requests import Response
from time import sleep
from stpstone._config.global_slots import YAML_WW_WORLD_GOV_BONDS, YAML_WW_RATINGS_AGENCIES
from stpstone.utils.cals.handling_dates import DatesBR
from stpstone.ingestion.abc.requests import ABCRequests
from stpstone.utils.webdriver_tools.playwright_wd import PlaywrightScraper
from stpstone.utils.parsers.dicts import HandlingDicts
from stpstone.utils.parsers.lists import ListHandler
from stpstone.utils.parsers.numbers import NumHandler


class WorldGovBonds(ABCRequests):

    def __init__(
        self,
        session: Optional[Session] = None,
        dt_ref: datetime = DatesBR().sub_working_days(DatesBR().curr_date, 1),
        cls_db: Optional[Session] = None,
        logger: Optional[Logger] = None,
        token: Optional[str] = None,
        list_slugs: Optional[List[str]] = None
    ) -> None:
        super().__init__(
            dict_metadata=YAML_WW_WORLD_GOV_BONDS,
            session=session,
            dt_ref=dt_ref,
            cls_db=cls_db,
            logger=logger,
            token=token,
            list_slugs=list_slugs
        )
        self.session = session
        self.dt_ref = dt_ref
        self.cls_db = cls_db
        self.logger = logger
        self.list_slugs = list_slugs

    def treat_list_td(self, list_td: List[str]) -> List[str]:
        """
        Processes the list of table data cells to insert "N/A" for countries without ratings.
        
        The rule is: After a country name, if the next item is a numeric value (like '9.88%' or '0.0988'),
        insert "N/A" between them to represent missing rating.
        
        Args:
            list_td: List of strings representing table data cells
            
        Returns:
            Processed list with "N/A" inserted where needed
        """
        list_ratings_agencies = ListHandler().extend_lists(
            *YAML_WW_RATINGS_AGENCIES["credit_ratings"].values())
        list_ = []
        i = 0
        n = len(list_td)
        while i < n:
            item_curr = list_td[i]
            item_processed = NumHandler().transform_to_float(list_td[i], int_precision=6)
            list_.append(item_processed)
            bl_country_name = (
                isinstance(item_curr, str) and
                not any(c.isdigit() for c in item_curr) and 
                not item_curr.endswith('%') and 
                item_curr != "N/A" and 
                item_curr not in list_ratings_agencies
            )
            if bl_country_name:
                if i + 1 < n:
                    next_item = list_td[i+1]
                    bl_next_numeric = (
                        (isinstance(next_item, str) and next_item.endswith('%')) 
                        or (isinstance(NumHandler().transform_to_float(next_item, int_precision=6), 
                                       (int, float)))
                    )
                    if bl_next_numeric:
                        list_.append("N/A")
            i += 1
        return list_

    def req_trt_injection(self, resp_req: Response) -> Optional[pd.DataFrame]:
        source = self.get_query_params(resp_req.url, "source")
        list_th = list(YAML_WW_WORLD_GOV_BONDS[source]["dtypes"].keys())
        scraper = PlaywrightScraper(
            bl_headless=True,
            int_default_timeout=100_000
        )
        with scraper.launch():
            if scraper.navigate(resp_req.url):
                list_td = scraper.get_list_data(
                    YAML_WW_WORLD_GOV_BONDS[source]["xpaths"]["list_td"],
                    selector_type="xpath"
                )
        list_td = self.treat_list_td(list_td)
        if source == "sovereign_cds": list_td = list_td[:-1]
        list_ser = HandlingDicts().pair_headers_with_data(list_th, list_td)
        return pd.DataFrame(list_ser)


### STREAMING PIPELINES ###

from typing import List, Callable


def streamingpipeline(generator, functions:List[Callable]):
    """
    Processes a generator stream through a sequence of functions
    Examples of usage:
        Real-time data processing, logs, market data
    Args:
        generator (iterable): A data stream
        functions (List[Callable]): A list of functions to apply
    Yields:
        Processed elements from the stream
    """
    for data in generator:
        for func in functions:
            data = func(data)
        yield data


if __name__ == '__main__':
    def to_uppercase(text): return text.upper()
    def add_exclamation(text): return text + '!'

    stream = iter(['hello', 'world'])
    processed_stream = streamingpipeline(stream, [to_uppercase, add_exclamation])

    for item in processed_stream:
        print(item)



import pandas as pd
from datetime import datetime
from typing import Optional
from sqlalchemy.orm import Session, Dict
from logging import Logger
from requests import Response
from stpstone._config.global_slots import YAML_B3_UP2DATA_VOLUMES_TRD
from stpstone.utils.cals.handling_dates import DatesBR
from stpstone.utils.connections.netops.proxies.managers.free import YieldFreeProxy
from stpstone.ingestion.abc.requests import ABCRequests


class ExchVolumesTrdBR(ABCRequests):

    def __init__(
        self,
        session: Optional[Session] = None,
        dt_ref:datetime=DatesBR().sub_working_days(DatesBR().curr_date, 1),
        dict_headers:Optional[Dict[str, str]]=None,
        dict_payload:Optional[Dict[str, str]]=None,
        cls_db:Optional[Session]=None,
        logger:Optional[Logger]=None
    ) -> None:
        self.token = self.access_token
        super().__init__(
            dict_metadata=YAML_B3_UP2DATA_VOLUMES_TRD,
            session=session,
            dt_ref=dt_ref,
            dict_headers=dict_headers,
            dict_payload=dict_payload,
            cls_db=cls_db,
            logger=logger
        )

    def trt_injection(self, resp_req:Response) -> Optional[pd.DataFrame]:
        return None



import pytz
import locale
import time
import businesstimedelta
import pandas as pd
from datetime import date, datetime, time, timedelta, timezone
from typing import List, Optional, Tuple, Union
from dateutil.relativedelta import relativedelta
from more_itertools import unique_everseen
from workalendar.core import SAT, SUN
from stpstone.transformations.validation.metaclass_type_checker import \
    TypeChecker
from stpstone.utils.cals.br_bzdays import BrazilBankCalendar
from stpstone.utils.parsers.str import StrHandler


class DatesBR(BrazilBankCalendar, metaclass=TypeChecker):

    def build_date(self, year: int, month: int, day: int) -> date:
        return date(year=year, month=month, day=day)

    def build_datetime(
        self, year: int, month: int, day: int, hour: int, minute: int, second: int
    ) -> datetime:
        return datetime(
            year=year, month=month, day=day, hour=hour, minute=minute, second=second
        )

    def date_to_datetime(
        self, date: date, bl_crop_time: bool = True, bl_tinestamp: bool = True
    ) -> datetime:
        datetime_ = datetime.combine(date, datetime.min.time())
        if bl_tinestamp == True:
            datetime_ = datetime_.timestamp_dt()
        if bl_crop_time == True:
            return int(datetime_)
        else:
            return datetime_

    def to_integer(self, dt_time: datetime) -> int:
        return 10000 * dt_time.year + 100 * dt_time.month + dt_time.day

    def excel_float_to_date(self, float_excel_date: float) -> date:
        return datetime.fromordinal(date(1900, 1, 1).toordinal() + float_excel_date - 2)

    def excel_float_to_datetime(self, float_excel_date):
        return datetime.fromordinal(
            datetime(1900, 1, 1).toordinal() + float_excel_date - 2
        )

    def check_is_date(self, dt_: datetime) -> bool:
        return isinstance(dt_, date)

    def str_date_to_datetime(self, date_str: str, format_input: str = "DD/MM/YYYY") -> datetime:
        """
        String date to datetime
        Args:
            date_str (str): date in string format
            format_input (str): output format - valid formats: 'DD/MM/YYYY', 'YYYY-MM-DD',
                'YYMMDD', 'DDMMYY', 'DDMMYYYY', 'DD/MM/YY'
        Returns:
            datetime
        """
        if format_input == "DD/MM/YYYY":
            return date(int(date_str[-4:]), int(date_str[3:5]), int(date_str[0:2]))
        elif format_input == "DDMMYY":
            return date(
                int("20" + date_str[-2:]), int(date_str[2:4]), int(date_str[0:2])
            )
        elif format_input == "DDMMYYYY":
            return date(int(date_str[-4:]), int(date_str[2:4]), int(date_str[0:2]))
        elif format_input == "YYYYMMDD":
            return date(int(date_str[:4]), int(date_str[4:6]), int(date_str[6:]))
        elif format_input == "YYYY-MM-DD":
            return date(int(date_str[0:4]), int(date_str[5:7]), int(date_str[-2:]))
        elif format_input == "MM-DD-YYYY":
            return date(int(date_str[-4:]), int(date_str[0:2]), int(date_str[3:5]))
        elif format_input == "YYMMDD":
            return date(
                int("20" + date_str[0:2]), int(date_str[2:4]), int(date_str[-2:])
            )
        elif format_input == "DD/MM/YY":
            return date(
                int("20" + date_str[-2:]), int(date_str[3:5]), int(date_str[0:2])
            )
        elif format_input == "DD.MM.YY":
            return date(
                int("20" + date_str[-2:]), int(date_str[3:5]), int(date_str[0:2])
            )
        else:
            raise Exception(f"Not a valid date format {format_input}")

    def list_wds(self, dt_start: datetime, dt_end: datetime, int_wd_bef: int) -> List[int]:
        """
        List of working days between two dates
        Args:
            dt_start (datetime): start date
            dt_end (datetime): end date
            int_wd_bef (int): number of working days before the start date
        Returns:
            List[int]
        """
        date_ref = self.sub_working_days(self.dt_(), int_wd_bef)
        wd_inf = self.get_working_days_delta(dt_start, date_ref)
        wd_sup = self.get_working_days_delta(dt_end, date_ref)
        return list(range(wd_sup, wd_inf + 1))

    def list_wds(self, dt_start: str, dt_end: str) -> Union[List[datetime], List[str]]:
        """
        List of working days between two dates
        Args:
            dt_start (datetime): start date
            dt_end (datetime): end date
        Returns:
            List[datetime]
        """
        list_wds = list()
        for x in range(int((dt_end - dt_start).days) + 1):
            list_wds.append(
                super().find_following_working_day(day=dt_start + timedelta(days=x))
            )
        return list(unique_everseen(list_wds))

    def list_cds(self, dt_start: datetime, dt_end: datetime) -> List[datetime]:
        """
        List of calendar days between two dates
        Args:
            dt_start (datetime): start date
            dt_end (datetime): end date
            format_data (str): format date
        Returns:
            List[datetime]
        """
        list_wds = list()
        for x in range(int((dt_end - dt_start).days)):
            list_wds.append(dt_start + timedelta(days=x))
        return list(unique_everseen(list_wds))

    def list_years(self, dt_start: datetime, dt_end: datetime) -> List[int]:
        """
        List of years between two dates
        Args:
            dt_start (datetime): start date
            dt_end (datetime): end date
        Returns:
            List[int]
        """
        list_years = list()
        for x in range(int((dt_end - dt_start).days)):
            list_years.append((dt_start + timedelta(days=x)).year)
        return list(unique_everseen(list_years))

    @property
    def curr_date(self) -> date:
        return date.today()

    @property
    def curr_time(self) -> datetime:
        return datetime.now().time()

    def curr_date_time(
        self, bl_timestamp: bool = False, bl_crop_time: bool = False
    ) -> Union[int, datetime]:
        if bl_timestamp == True:
            datetime_ = datetime.now().timestamp_dt()
        else:
            datetime_ = datetime.now()
        if bl_crop_time == True:
            return int(datetime_)
        else:
            return datetime_

    def testing_dates(self, dt_start: datetime, dt_end: datetime) -> bool:
        """
        Test if dt_end is greater than dt_start
        Args:
            dt_start (datetime): start date
            dt_end (datetime): end date
        Returns:
            Boolean
        """
        if int((dt_end - dt_start).days) >= 0:
            return True
        else:
            return False

    def year_number(self, dt_: Union[date, datetime]) -> int:
        return int(dt_.strftime("%Y"))

    def day_number(self, dt_: Union[date, datetime]) -> int:
        return int(dt_.strftime("%d"))

    def month_name(
        self,
        dt_: Union[date, datetime],
        bl_abbrv: bool = False,
        local_zone: str = "pt-BR",
    ) -> str:
        """
        Name of the month in the local language
        Args:
            dt_ (date): date
            bl_abbrv (bool): abbreviation
            local_zone (str): local zone
        Returns:
            str
        """
        locale.setlocale(locale.LC_TIME, local_zone)
        if bl_abbrv == True:
            return dt_.strftime("%b")
        else:
            return dt_.strftime("%B")

    def dates_inf_sup_month(self, dt_, last_month_year=12) -> Tuple[date, date]:
        year = self.year_number(dt_)
        month = self.month_number(dt_)
        day = 1
        dt_start = self.find_working_day(self.build_date(year, month, day))
        if month < last_month_year:
            dt_end = self.sub_working_days(self.build_date(year, month + 1, day), 1)
        else:
            dt_end = self.sub_working_days(self.build_date(year + 1, 1, day), 1)
        # returning dates
        return dt_start, dt_end

    def month_number(self, dt_: datetime, bl_month_mm: bool = False) -> Union[int, str]:
        if bl_month_mm == False:
            return int(dt_.strftime("%m"))
        else:
            return dt_.strftime("%m")

    def week_name(
        self, dt_: datetime, bl_abbrv: bool = False, local_zone: str = "pt-BR"
    ) -> str:
        """
        Name of the weekday in the local language
        Args:
            dt_ (date): date
            bl_abbrv (bool): abbreviation
            local_zone (str): local zone
        Returns:
            str
        """
        locale.setlocale(locale.LC_TIME, local_zone)
        if bl_abbrv == True:
            return dt_.strftime("%a")
        else:
            return dt_.strftime("%A")

    def week_number(self, dt_: datetime) -> str:
        return dt_.strftime("%w")

    def find_working_day(self, dt_: datetime) -> datetime:
        return self.add_working_days(self.sub_working_days(dt_, 1), 1)

    def nth_weekday_month(
        self,
        dt_start: datetime,
        dt_end: datetime,
        int_weekday: int,
        nth_rpt: int,
        format_output: str = "DD/MM/YYYY",
        int_days_week: int = 7,
    ) -> List[datetime]:
        """
        Get nth weekday of month
        Args:
            dt_start (datetime): start date
            dt_end (datetime): end date
            int_weekday (int): weekday number
            nth_rpt (int): nth repetition
            format_output (str): format output
            int_days_week (int): number of days in a week
        Returns:
            List[datetime]
        """
        list_wds = self.list_wds(dt_start, dt_end, format_output)
        return [
            self.add_working_days(self.sub_working_days(d, 1), 1)
            for d in list_wds
            if (
                self.week_number(d) == int_weekday
                and d.day >= (nth_rpt * int_days_week - int_days_week)
                and d.day <= (nth_rpt * int_days_week)
            )
        ]

    def delta_calendar_days(self, dt_start: datetime, dt_end: datetime) -> int:
        return (dt_end - dt_start).days

    def add_months(self, dt_, int_months) -> datetime:
        return dt_ + relativedelta(months=int_months)

    def add_calendar_days(self, original_date, days_to_add):
        return original_date + timedelta(days=days_to_add)

    def delta_working_hours(
        self,
        timestamp_inf: str,
        timestamp_sup: str,
        int_hour_start_office: int = 8,
        int_hour_sup_office: int = 18,
        int_hour_start_lunch: int = 0,
        int_hour_sup_lunch: int = 0,
        list_wds: List[int] = [0, 1, 2, 3, 4],
    ) -> int:
        """
        Calculate the number of working hours between two timestamps
        Args:
            timestamp_inf (str): start timestamp_dt
            timestamp_sup (str): end timestamp_dt
            int_hour_start_office (int): start hour office
            int_hour_sup_office (int): end hour office
            int_hour_start_lunch (int): start hour lunch
            int_hour_sup_lunch (int): end hour lunch
            list_wds (List[int]): list of working days
        Returns:
            int
        References: https://pypi.org/project/businesstimedelta/
        """
        # timestamp_dt convertation to datetime
        y_inf, mt_inf, d_inf = (
            int(timestamp_inf.split(" ")[0].split("-")[0]),
            int(timestamp_inf.split(" ")[0].split("-")[1]),
            int(timestamp_inf.split(" ")[0].split("-")[2]),
        )
        h_inf, m_inf, s_inf = (
            int(timestamp_inf.split(" ")[1].split(":")[0]),
            int(timestamp_inf.split(" ")[1].split(":")[1]),
            int(timestamp_inf.split(" ")[1].split(":")[2]),
        )
        y_sup, mt_sup, d_sup = (
            int(timestamp_sup.split(" ")[0].split("-")[0]),
            int(timestamp_sup.split(" ")[0].split("-")[1]),
            int(timestamp_sup.split(" ")[0].split("-")[2]),
        )
        h_sup, m_sup, s_sup = (
            int(timestamp_sup.split(" ")[1].split(":")[0]),
            int(timestamp_sup.split(" ")[1].split(":")[1]),
            int(timestamp_sup.split(" ")[1].split(":")[2]),
        )
        timestamp_inf = datetime(y_inf, mt_inf, d_inf, h_inf, m_inf, s_inf)
        timestamp_sup = datetime(y_sup, mt_sup, d_sup, h_sup, m_sup, s_sup)
        # dict of holidays
        dict_holidays_raw = dict()
        for y in range(timestamp_inf.year, timestamp_sup.year + 1):
            dict_holidays_raw[y] = self.holidays(y)
        dict_holidays_trt = dict()
        for k, v in dict_holidays_raw.items():
            for t in v:
                dict_holidays_trt[t[0]] = t[1]
        # office hours for working days
        workday = businesstimedelta.WorkDayRule(
            start_time=time(int_hour_start_office),
            end_time=time(int_hour_sup_office),
            list_wds=list_wds,
        )
        lunchbreak = businesstimedelta.LunchTimeRule(
            start_time=time(int_hour_start_lunch),
            end_time=time(int_hour_sup_lunch),
            list_wds=list_wds,
        )
        holidays = businesstimedelta.HolidayRule(dict_holidays_trt)
        businesshrs = businesstimedelta.Rules([workday, lunchbreak, holidays])
        # output
        return businesshrs.difference(timestamp_inf, timestamp_sup).timedelta

    def last_wd_years(self, list_years: List[int]) -> List[datetime]:
        """
        Last days of years
        Args:
            list_years (List[int]): list of years
        Returns:
            List[datetime]
        """
        return [self.sub_working_days(datetime(y + 1, 1, 1), 1) for y in list_years]

    def add_holidays_not_considered_anbima(
        self,
        dt_start: datetime,
        dt_end: datetime,
        list_last_week_year_day,
        local_zone="pt-BR",
        list_holidays_not_considered: List[str] = ["25/01"],
        list_dates_not_considered: List[str] = ["05/03/2025", "18/02/2026"],
        list_non_bzdays_week: List[str] = ["sábado", "domingo"],
    ):
        """
        Add holidays not considered by ANBIMA
        Args:
            dt_start (datetime): start date
            dt_end (datetime): end date
            list_last_week_year_day (List[datetime]): list of last days of years
            local_zone (str): locale zone
            list_holidays_not_considered (List[str]): list of holidays not considered
            list_dates_not_considered (List[str]): list of dates not considered
            list_non_bzdays_week (List[str]): list of non business days of the week
        Returns:
            int
        """
        locale.setlocale(locale.LC_TIME, local_zone)
        return len(
            [
                d
                for d in self.list_calendar_days(dt_start, dt_end)
                if (
                    d.strftime("%d/%m") in list_holidays_not_considered
                    and not self.week_name(d) in list_non_bzdays_week
                    or d in list_last_week_year_day
                    or d.strftime("%d/%m/%Y") in list_dates_not_considered
                )
            ]
        )

    def unix_timestamp_to_datetime(
        self, unix_timestamp: Union[float, int], str_tz: str = "UTC"
    ) -> datetime:
        tz_obj = timezone.utc if str_tz == "UTC" else timezone(str_tz)
        return datetime.fromtimestamp(unix_timestamp, tz=tz_obj)

    def unix_timestamp_to_date(
        self, unix_timestamp: Union[float, int], str_tz: str = "UTC"
    ) -> datetime:
        tz_obj = timezone.utc if str_tz == "UTC" else timezone(str_tz)
        return datetime.fromtimestamp(unix_timestamp, tz=tz_obj).date()

    def iso_to_unix_timestamp(self, iso_timestamp: str) -> int:
        dt_ = datetime.fromisoformat(iso_timestamp)
        dt_utc = dt_.astimezone(timezone.utc)
        return dt_utc.timestamp()

    def datetime_to_unix_timestamp(self, dt_: Union[date, datetime, time]) -> int:
        """
        Convert a datetime/date/time object to a Unix timestamp (seconds since epoch).

        Args:
            dt_ (Union[date, datetime, time]): Date, datetime or time object to convert

        Returns:
            int: Unix timestamp (seconds since 1970-01-01 00:00:00 UTC)
        """
        # If input is time, combine with today's date
        if isinstance(dt_, time):
            dt_ = datetime.combine(date.today(), dt_)
        # If input is date, convert to datetime at midnight
        elif isinstance(dt_, date) and not isinstance(dt_, datetime):
            dt_ = datetime.combine(dt_, time.min)

        # If datetime is timezone-naive, assume local timezone
        if dt_.tzinfo is None:
            dt_ = dt_.astimezone()  # Convert to local timezone

        # Convert to UTC and return timestamp
        return int(dt_.astimezone(timezone.utc).timestamp())

    def timestamp_to_date(
        self,
        timestamp: Union[str, int],
        substring_datetime: Optional[str] = "T",
        format_output: str = "YYYY-MM-DD",
    ) -> datetime:
        if substring_datetime == None:
            return datetime.fromtimestamp(int(timestamp) / 1000, tz=timezone.utc)
        return self.str_date_to_datetime(
            StrHandler().get_string_until_substr(str(timestamp), substring_datetime),
            format_output,
        )

    def timestamp_to_datetime(
        self, timestamp_dt: datetime, bl_return_from_utc: bool = False
    ) -> Union[datetime, str]:
        if bl_return_from_utc == True:
            return pd.to_datetime(timestamp_dt, unit="s", utc=True).tz_convert(
                "America/Sao_Paulo"
            )
        else:
            return pd.to_datetime(timestamp_dt, unit="s", utc=True).strftime("%Y%m%d")

    @property
    def current_timestamp_string(self, format_output: str = "%Y%m%d_%H%M%S") -> str:
        return self.curr_date_time().strftime(format_output)

    @property
    def utc_log_ts(self) -> datetime:
        return datetime.now(timezone.utc)

    def utc_from_dt(self, dt_):
        dt_ = datetime.combine(dt_, datetime.min.time())
        return pytz.utc.localize(dt_)

    def month_year_string(
        self,
        dt_: str,
        format_input: str = "%b/%Y",
        format_output: str = "%Y-%m",
        bl_dtbr: bool = True,
    ):
        if bl_dtbr == True:
            month_mapping = {
                "JAN": "01",
                "FEB": "02",
                "MAR": "03",
                "APR": "04",
                "MAY": "05",
                "JUN": "06",
                "JUL": "07",
                "AUG": "08",
                "SEP": "09",
                "OCT": "10",
                "NOV": "11",
                "DEC": "12",
            }
            month_abbr, year = dt_.split("/")
            month = month_mapping[month_abbr.upper()]
            return f"{year}-{month}"
        else:
            return datetime.strptime(dt_, format_input).strftime(format_output)



from stpstone.transformations.validation.metaclass_type_checker import TypeChecker


class BinaryMultiplier(metaclass=TypeChecker):

    def __init__(self, a: int, b: int) -> None:
        """
        Initialize the Binary Multiplier with two binary numbers.
        Args:
            a (int): First binary number
            b (int): Second binary number
        Returns:
            None
        """
        self.a = a
        self.b = b

    def multiply(self) -> int:
        """
        Multiply two binary numbers and return the result.

        The multiplication is performed by shifting the bits of the first number
        (self.a) to the left by the bit position of the second number (self.b) and
        adding the results.

        Args:
            None

        Returns:
            int: The result of multiplying the two binary numbers.
        """
        result = 0
        for i in range(8):
            # check if the i-th bit of b is set
            if (self.b >> i) & 1:
                # shift a left by i and add to result
                result += (self.a << i)
        return result


# example usage of the binary multiplier
if __name__ == "__main__":
    multiplier = BinaryMultiplier(0b1100, 0b1010)
    product = multiplier.multiply()
    print(f"Binary Multiplier (1100, 1010) => Product: {bin(product)}")  # output: Product in binary



import numpy as np
from typing import Any


class Stack:

    def __init__(self, initial_capacity: int = 10) -> None:
        self._items = np.empty(initial_capacity, dtype=object)
        self._size = 0

    @property
    def is_empty(self) -> bool:
        return self._size == 0

    def push(self, item: Any) -> None:
        if self._size == len(self._items):
            self._resize(2 * len(self._items))
        self._items[self._size] = item
        self._size += 1

    def pop(self) -> Any:
        if self.is_empty:
            raise IndexError("Pop from an empty stack")
        self._size -= 1
        item = self._items[self._size]
        self._items[self._size] = None
        return item

    @property
    def peek(self) -> Any:
        if self.is_empty:
            raise IndexError("Peek from an empty stack")
        return self._items[self._size - 1]

    @property
    def size(self) -> int:
        return self._size

    def _resize(self, new_capacity: int) -> None:
        """
        Private method to resize the underlying array of the stack to the new capacity.

        Args:
            new_capacity (int): New capacity of the stack

        Returns:
            None
        """
        new_items = np.empty(new_capacity, dtype=object)
        new_items[:self._size] = self._items[:self._size]
        self._items = new_items



import pandas as pd
from datetime import datetime
from typing import Optional, List, Any, Tuple
from sqlalchemy.orm import Session
from logging import Logger
from requests import Response
from time import sleep
from lxml.html import HtmlElement
from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.common.by import By
from selenium.common.exceptions import NoSuchElementException
from selenium.common.exceptions import TimeoutException
from stpstone._config.global_slots import YAML_US_ETFDB_VETTAFI
from stpstone.utils.cals.handling_dates import DatesBR
from stpstone.ingestion.abc.requests import ABCRequests
from stpstone.utils.parsers.folders import DirFilesManagement
from stpstone.utils.parsers.dicts import HandlingDicts
from stpstone.utils.parsers.str import StrHandler
from stpstone.utils.loggs.create_logs import CreateLog


class EtfDBVettaFi(ABCRequests):

    def __init__(
        self,
        session: Optional[Session] = None,
        dt_ref: datetime = DatesBR().sub_working_days(DatesBR().curr_date, 1),
        cls_db: Optional[Session] = None,
        logger: Optional[Logger] = None,
        token: Optional[str] = None,
        list_slugs: Optional[List[str]] = None,
        int_wait_load_seconds: int = 10,
        bl_headless: bool = False,
        bl_incognito: bool = False
    ) -> None:
        super().__init__(
            dict_metadata=YAML_US_ETFDB_VETTAFI,
            session=session,
            dt_ref=dt_ref,
            cls_db=cls_db,
            logger=logger,
            token=token,
            list_slugs=list_slugs,
            int_wait_load_seconds=int_wait_load_seconds,
            bl_headless=bl_headless,
            bl_incognito=bl_incognito
        )
        self.session = session
        self.dt_ref = dt_ref
        self.cls_db = cls_db
        self.logger = logger
        self.token = token
        self.list_slugs = list_slugs

    def td_th_parser(self, root: HtmlElement) -> pd.DataFrame:
        list_ser = list()
        for i in range(1, 50):
            try:
                el_tr = root.find_element(getattr(By, "XPATH"),
                    YAML_US_ETFDB_VETTAFI["reits"]["xpaths"]["list_tr"].format(i)
                )
                list_ser.append({
                    "SYMBOL": el_tr.find_element(getattr(By, "XPATH"), "./td[1]/a").text.strip(),
                    "HOLDING": el_tr.find_element(getattr(By, "XPATH"), "./td[2]").text.strip(),
                    "WEIGHT": float(el_tr.find_element(getattr(By, "XPATH"), "./td[3]").text.strip()\
                        .replace("%", "")) / 100.0
                })
            except (NoSuchElementException, TimeoutException):
                break
        if len(list_ser) == 0:
            return [{"SYMBOL": "ERROR", "HOLDING": "ERROR", "WEIGHT": 0.0}]
        return list_ser

    def req_trt_injection(self, web_driver: WebDriver) -> Optional[pd.DataFrame]:
        list_ser = self.td_th_parser(web_driver)
        return pd.DataFrame(list_ser)



import pandas as pd
from datetime import datetime
from typing import Optional, List, Any, Dict
from sqlalchemy.orm import Session
from logging import Logger
from requests import Response
from time import sleep
from stpstone._config.global_slots import YAML_YAHII_OHTERS
from stpstone.utils.cals.handling_dates import DatesBR
from stpstone.ingestion.abc.requests import ABCRequests
from stpstone.utils.parsers.numbers import NumHandler
from stpstone.utils.parsers.html import HtmlHandler
from stpstone.utils.parsers.folders import DirFilesManagement
from stpstone.utils.parsers.dicts import HandlingDicts
from stpstone.utils.parsers.str import StrHandler
from stpstone.utils.loggs.create_logs import CreateLog


class YahiiOthersBR(ABCRequests):

    def __init__(
        self,
        session: Optional[Session] = None,
        int_delay_seconds: int = 20,
        dt_ref: datetime = DatesBR().sub_working_days(DatesBR().curr_date, 1),
        cls_db: Optional[Session] = None,
        logger: Optional[Logger] = None,
        token: Optional[str] = None,
        list_slugs: Optional[List[str]] = None
    ) -> None:
        super().__init__(
            dict_metadata=YAML_YAHII_OHTERS,
            session=session,
            dt_ref=dt_ref,
            cls_db=cls_db,
            logger=logger,
            token=token,
            list_slugs=list_slugs, 
            int_delay_seconds=int_delay_seconds,
        )
        self.session = session
        self.dt_ref = dt_ref
        self.cls_db = cls_db
        self.logger = logger
        self.token = token
        self.list_slugs = list_slugs
        self.int_delay_seconds = int_delay_seconds
        self.year_yy = self.dt_ref.strftime("%y")

    def td_th_parser(self, list_td: List[Any], list_headers: List[str], source: str) \
        -> pd.DataFrame:
        if source == "min_wage":
            for i, td in enumerate(list_td):
                if "DECRETO" not in td and "$" in td:
                    list_td[i] = float(td.replace("NCz$ ", "").replace("NCr$ ", "")\
                        .replace("CR$", "").replace("R$", "").replace("Cr$", "").replace("NCz$", "")\
                        .replace("$000", "").replace(".", "")\
                        .replace(",", ".").replace("Cz$ ", "").strip())
                elif "DECRETO" not in td and len(td) >= 8 and "." in td:
                    list_td[i] = td.replace(" (Abono)", "").replace(" (URV) – (Real)", "")\
                        .replace(" (Cruzeiro Real)", "").replace(" (Cruzeiro)", "")\
                        .replace(" (Cruzado)", "").replace(" (Cruzeiro Novo)", "")\
                        .replace(" (Cruzado Novo)", "").replace(" (Réis)", "")
            list_ser = HandlingDicts().pair_headers_with_data(
                list_headers,
                list_td
            )
            df_ = pd.DataFrame(list_ser)
            df_ = df_[df_["DATA"] != "REVOGADA"]
            df_["DATA"] = [DatesBR().str_date_to_datetime(d, "DD.MM.YY") for d in df_["DATA"]]
            df_ = df_[(df_["DATA"] >= DatesBR().build_date(2000, 1, 1)) & (df_["DATA"] <= self.dt_ref)]
        elif source == "inss_contribution":
            list_td = [
                x.replace("\r\n                 ", "").strip().replace("%", "")\
                    .replace(",01,01", "").replace("Até  4.190,84 ", "De 4.190,84 ").strip()\
                for x in list_td
            ]
            list_td = [
                float(x.replace(",", ".")) / 100.0 if NumHandler().is_numeric(x.replace(",", ".")) \
                else x for x in list_td
            ]
            list_ser = HandlingDicts().pair_headers_with_data(
                list_headers,
                list_td
            )
            df_ = pd.DataFrame(list_ser)
            df_["SALARIO_INF"] = [
                StrHandler().get_between(StrHandler().remove_diacritics(x.lower()), "de", "ate")\
                    .strip() if "de " in x.lower() and "acima de " not in x.lower() else \
                StrHandler().get_after(StrHandler().remove_diacritics(x.lower().strip()), "acima de ")\
                    .strip() if "acima de " in StrHandler().remove_diacritics(x.lower().strip()) \
                    else 0.0 for x in df_["SALARIO_CONTRIBUICAO"]
            ]
            df_["SALARIO_SUP"] = [
                StrHandler().get_after(StrHandler().remove_diacritics(x.lower()), "ate ")\
                    .strip() if "ate " in StrHandler().remove_diacritics(x.lower()) else 0.0 \
                for x in df_["SALARIO_CONTRIBUICAO"]
            ]
            df_["SALARIO_INF"] = pd.to_numeric(
                df_["SALARIO_INF"].str.replace(".", "").str.replace(",", "."), 
                errors="coerce"
            )
            df_["SALARIO_SUP"] = pd.to_numeric(
                df_["SALARIO_SUP"].str.replace(".", "").str.replace(",", "."), 
                errors="coerce"
            )
            df_["SALARIO_INF"] = df_["SALARIO_INF"].fillna(0.0)
            df_["SALARIO_SUP"] = df_["SALARIO_SUP"].fillna(1_000_000_000)
        elif source in ["daily_usdbrl", "daily_eurbrl"]:
            list_td = [x.replace(",", ".") for x in list_td \
                       if (StrHandler().match_string_like(x, "*/*/*") and len(x) == 10) \
                        or "," in x or "." in x \
                        or (StrHandler().match_string_like(x, "*/*") and len(x) == 9 \
                            and StrHandler().has_no_letters(x))]
            list_ser = HandlingDicts().pair_headers_with_data(
                list_headers,
                list_td
            )
            df_ = pd.DataFrame(list_ser)
        return df_

    def req_trt_injection(self, resp_req: Response) -> Optional[pd.DataFrame]:
        root = HtmlHandler().lxml_parser(resp_req)
        source = self.get_query_params(resp_req.url, "source").lower()
        list_td = [
            x.text.strip() for x in HtmlHandler().lxml_xpath(
                root, YAML_YAHII_OHTERS[source]["xpaths"]["list_td"]
            )
            if x.text is not None
        ]
        if source == "min_wage":
            list_headers = ["DISPOSITIVO_LEGAL", "DATA", "VALOR"]
        elif source == "inss_contribution":
            list_headers = ["SALARIO_CONTRIBUICAO", "ALIQUOTA_RECOLHIMENTO_INSS"]
        elif source in ["daily_usdbrl", "daily_eurbrl"]:
            list_headers = ["DATA", "COMPRA", "VENDA"]
        else:
            raise ValueError(f"Source {source} not implemented.")
        return self.td_th_parser(list_td, list_headers, source)


### MODULE TO HANDLE SENDGRID X PYTHON INTEGRATION ###

from sendgrid import SendGridAPIClient
from sendgrid.helpers.mail import Mail


class SendGrid:

    def send_email(self, str_sender, list_recipients, list_cc, subject, html_body, token):
        """
        DOCSTRING: SEND EMAIL FROM SENDGRID
        INPUTS: SENDER STR, LIST OF RECIPIENTS STRINGS, SUBJECT, HTML BODY (STARTING AND ENDING WITH DOUBLE QUOTES, USING ONLY SINGLE QUOTES INSIDE IT) AND TOKEN
        OUPUTS: STATUS OF ACCOMPLISHMENT
        """
        message = Mail(
            str_sender=str_sender,
            to_emails=list_recipients,
            mail_cc=list_cc,
            subject=subject,
            html_content=html_body)
        try:
            sg = SendGridAPIClient(token)
            return sg.send(message)
        except Exception as e:
            raise Exception('Error sending email from sendgrid: {}'.format(e))



import pandas as pd
from pymongo import MongoClient
from logging import Logger
from typing import Optional
from __future__ import annotations
from stpstone.utils.loggs.create_logs import CreateLog


class MongoConn:

    _instance: MongoConn | None = None

    def __new__(cls:type[MongoConn], *args, **kwargs) -> MongoConn:
        """
        DOCSTRING: ENSURES THAT ONLY ONE INSTANCE OF THE CLASS IS CREATED (SINGLETON PATTERN)
        INPUTS: ARGS, KWARGS
        OUTPUS: MONGOCONN INSTANCE
        """
        if not cls._instance:
            cls._instance = super().__new__(cls)
            cls._instance._client = None
            cls._instance.db = None
            cls._instance._collection = None
        return cls._instance

    def __init__(self, str_host:str='localhost', int_port:int=27017, str_dbname:str='test', str_collection:str='data',
                 logger:Optional[Logger]=None) -> None:
        """
        DOCSTRING: INITIALIZES THE CONNECTION TO MONGODB
        INPUTS:
            - STR_HOST (STR): ADDRESS OF THE MONGODB SERVER
            - INT_PORT (INT): PORT OF THE MONGODB SERVER
            - STR_DBNAME (STR): NAME OF THE DATABASE
            - STR_COLLECTION (STR): NAME OF THE COLLECTION
            - LOGGER (OPTIONAL[LOGGER]): LOGGER INSTANCE FOR LOGGING
        OUTPUTS: -
        """
        if self._client is None:
            self.str_host = str_host
            self.int_port = int_port
            self.str_dbname = str_dbname
            self.str_collection = str_collection
            self.logger = logger
            self._connect

    @property
    def _connect(self) -> None:
        """
        DOCSRTING: CONNECTS TO THE MONGODB SERVER AND INITIALIZES THE DATABASE AND COLLECTION
        INPUTS: -
        OUTPUTS: -
        """
        self._client = MongoClient(self.str_host, self.int_port)
        self._db = self._client[self.str_dbname]
        self._collection = self._db[self.str_collection]

    def save_df(self, df_:pd.DataFrame) -> None:
        """
        DOCSTRING: SAVES A PANDAS DATAFRAME TO THE MONGODB COLLECTION
        INPUTS:
            df_ (pd.DataFrame): DATAFRAME TO BE SAVED
        OUTPUTS: -
        """
        if not isinstance(df_, pd.DataFrame):
            raise ValueError("THE PROVIDED DATA IS NOT A PANDAS DATAFRAME.")
        data = df_.to_dict(orient='records')
        try:
            self._collection.insert_many(data)
        except Exception as e:
            if self.logger is not None:
                CreateLog().info(self.logger, f'ERROR {e}, MONGODB INJECTION ABORTED')

    @property
    def close(self) -> None:
        """
        DOCSTRING: CLOSES THE CONNECTION TO THE MONGODB SERVER
        INPUTS: -
        OUTPUTS: -
        """
        if self._client:
            self._client.close()

    def __enter__(self) -> MongoConn:
        """
        DOCSTRING: RETURNS THE INSTANCE FOR USE IN A CONTEXT MANAGER
        INPUTS: -
        OUTPUTS: MONGOCONN INSTANCE
        """
        return self

    def __exit__(self, exc_type: Optional[type[BaseException]], exc_val: Optional[BaseException],
                 exc_tb: Optional[BaseException]) -> None:
        """
        DOCSTRING: CLOSES THE CONNECTION WHEN EXITING THE CONTEXT
        INPUTS:
            - EXC_TYPE: EXCEPTION TYPE
            - EXC_VAL: EXCEPTION VALUE
            - EXC_TB: TRACEBACK
        OUTPUTS: -
        """
        self.close



from requests import request
from typing import Dict, Union, List, Optional
from logging import Logger
from stpstone.utils.connections.netops.proxies.abc import ABCSession
from stpstone.utils.cals.handling_dates import DatesBR
from stpstone.utils.geography.ww import WWTimezones, WWGeography
from stpstone.utils.parsers.html import HtmlHandler
from stpstone.utils.parsers.str import StrHandler


class FreeProxyNet(ABCSession):

    def __init__(
        self,
        bl_new_proxy: bool = True,
        dict_proxies: Union[Dict[str, str], None] = None,
        int_retries: int = 10,
        int_backoff_factor: int = 1,
        bl_alive: bool = True,
        list_anonymity_value: List[str] = ["anonymous", "elite"],
        list_protocol: str = 'http',
        str_continent_code: Union[str, None] = None,
        str_country_code: Union[str, None] = None,
        bl_ssl: Union[bool, None] = None,
        float_min_ratio_times_alive_dead: Optional[float] = 0.02,
        float_max_timeout: Optional[float] = 600,
        bl_use_timer: bool = False,
        list_status_forcelist: List[int] = [429, 500, 502, 503, 504],
        logger: Optional[Logger] = None
    ) -> None:
        super().__init__(
            bl_new_proxy=bl_new_proxy,
            dict_proxies=dict_proxies,
            int_retries=int_retries,
            int_backoff_factor=int_backoff_factor,
            bl_alive=bl_alive,
            list_anonymity_value=list_anonymity_value,
            list_protocol=list_protocol,
            str_continent_code=str_continent_code,
            str_country_code=str_country_code,
            bl_ssl=bl_ssl,
            float_min_ratio_times_alive_dead=float_min_ratio_times_alive_dead,
            float_max_timeout=float_max_timeout,
            bl_use_timer=bl_use_timer,
            list_status_forcelist=list_status_forcelist,
            logger=logger
        )
        self.dates_br = DatesBR()
        self.ww_timezones = WWTimezones()
        self.ww_geography = WWGeography()
        self.html_parser = HtmlHandler()
        self.str_parser = StrHandler()

    @property
    def _available_proxies(self) -> List[Dict[str, Union[str, float]]]:
        list_ser = []
        payload = {}
        headers = {
            'accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7',
            'accept-language': 'en-US,en;q=0.9,pt;q=0.8,es;q=0.7',
            'cache-control': 'max-age=0',
            'if-modified-since': 'Sun, 06 Apr 2025 10:02:02 GMT',
            'priority': 'u=0, i',
            'sec-ch-ua': '"Google Chrome";v="135", "Not-A.Brand";v="8", "Chromium";v="135"',
            'sec-ch-ua-mobile': '?0',
            'sec-ch-ua-platform': '"Windows"',
            'sec-fetch-dest': 'document',
            'sec-fetch-mode': 'navigate',
            'sec-fetch-site': 'none',
            'sec-fetch-user': '?1',
            'upgrade-insecure-requests': '1',
            'user-agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/135.0.0.0 Safari/537.36',
            'Cookie': '_ga=GA1.1.2121373567.1743934016; _ga_F5HK5559Z2=GS1.1.1743937290.2.1.1743937312.0.0.0'
        }
        resp_req = request("GET", "https://free-proxy-list.net/", headers=headers, data=payload,
                           timeout=10)
        resp_req.raise_for_status()
        html_root = self.html_parser.lxml_parser(resp_req)
        el_trs = self.html_parser.lxml_xpath(
            html_root, '//*[@id="list"]/div/div[2]/div/table/tbody/tr')
        for el_tr in el_trs:
            str_ip = self.html_parser.lxml_xpath(el_tr, './td[1]')[0].text
            str_port = self.html_parser.lxml_xpath(el_tr, './td[2]')[0].text
            str_country_code = self.html_parser.lxml_xpath(el_tr, './td[3]')[0].text
            str_country_name = self.html_parser.lxml_xpath(el_tr, './/td[@class="hm"][1]')[0].text
            str_anonymity = self.html_parser.lxml_xpath(el_tr, './td[5]')[0].text
            str_anonymity = "elite" if self.str_parser.match_string_like(str_anonymity, "elite*") \
                else (
                    "anonymous" if self.str_parser.match_string_like(str_anonymity, "anonymous*") \
                    else "transparent"
                )
            str_protocol_code = self.html_parser.lxml_xpath(el_tr, './/td[@class="hx"]')[0].text
            str_protocol = "https" if self.str_parser.match_string_like(str_protocol_code, "yes*") \
                else "http"
            str_last_checked = self.html_parser.lxml_xpath(el_tr, './/td[@class="hm"][3]')[0].text
            obj_timezone = self.ww_timezones.get_timezones_by_country_code(str_country_code) if \
                str_country_code is not None else None
            str_timezone = ", ".join(obj_timezone) if isinstance(obj_timezone, (list, tuple, set)) \
                else (str(obj_timezone) if obj_timezone is not None else "Unknown")
            str_continent = self.ww_geography.get_continent_by_country_code(str_country_code) if \
                str_country_code is not None else "Unknown"
            str_continent_code = self.ww_geography.get_continent_code_by_country_code(
                str_country_code) if str_country_code is not None else "Unknown"
            list_ser.append({
                "protocol": str_protocol,
                "bl_alive": True,
                "status": "success",
                "alive_since": self.time_ago_to_ts_unix(str_last_checked),
                "anonymity": str_anonymity.lower(),
                "average_timeout": 1.0,
                "first_seen": self.time_ago_to_ts_unix(str_last_checked),
                "ip_data": "",
                "ip_name": "",
                "timezone": str_timezone,
                "continent": str_continent,
                "continent_code": str_continent_code,
                "country": str_country_name,
                "country_code": str_country_code if str_country_code is not None else "Unknown",
                "city": "",
                "district": "",
                "region_name": "",
                "zip": "",
                "bl_hosting": False,
                "isp": "",
                "latitude": 0.0,
                "longitude": 0.0,
                "organization": "",
                "proxy": True,
                "ip": str_ip,
                "port": str_port,
                "bl_ssl": True,
                "timeout": 1.0,
                "times_alive": 1,
                "times_dead": 0,
                "ratio_times_alive_dead": 1.0,
                "uptime": 1.0
            })
        return list_ser



class ColorIdentifier:
    """Identify if a hex color belongs to a specific color family (e.g., green, red)."""
    
    def __init__(self, hex_color: str):
        self.hex = hex_color.lstrip('#').lower()
        self.r = int(self.hex[0:2], 16)
        self.g = int(self.hex[2:4], 16)
        self.b = int(self.hex[4:6], 16)
    
    def bl_green(self, threshold: float = 1.5) -> bool:
        """Return True if green is dominant over red/blue by a threshold multiplier."""
        return (self.g > threshold * self.r) and (self.g > threshold * self.b)
    
    def bl_red(self, threshold: float = 1.5) -> bool:
        """Return True if red is dominant over green/blue."""
        return (self.r > threshold * self.g) and (self.r > threshold * self.b)
    
    def bl_blue(self, threshold: float = 1.5) -> bool:
        """Return True if blue is dominant over red/green."""
        return (self.b > threshold * self.r) and (self.b > threshold * self.g)
    
    def bl_white(self, min_intensity: int = 230) -> bool:
        """Return True if all RGB components are near maximum (e.g., #ffffff)."""
        return all(c >= min_intensity for c in [self.r, self.g, self.b])
    
    def bl_black(self, max_intensity: int = 25) -> bool:
        """Return True if all RGB components are near minimum (e.g., #000000)."""
        return all(c <= max_intensity for c in [self.r, self.g, self.b])
    
    def bl_gray(self, tolerance: int = 10) -> bool:
        """Return True if R ≈ G ≈ B within a tolerance range."""
        return (
            abs(self.r - self.g) <= tolerance and 
            abs(self.g - self.b) <= tolerance
        )

    def bl_yellow(self, threshold: float = 1.5) -> bool:
        """Return True if red+green dominate and blue is low."""
        return (self.r + self.g > threshold * self.b) and (self.b < 100)
    
    def bl_purple(self, threshold: float = 1.5) -> bool:
        """Return True if red+blue dominate and green is low."""
        return (self.r + self.b > threshold * self.g) and (self.g < 100)